/******************************************************************************
 *
 *    File Name:  clk_dlls.v
 *      Version:  2.5
 *         Date:  June 05, 2000
 * 
 *  Description:   clk_ddls contain 2 dlls to deskew FPGA and DDR SDRAM clocks. 
 * 
 *       	   dll_ext provides ddr_clk and ddr_clkb for the SDRAM and receives feedback
 *                 from the SDRAM
 * 
 *       	   dll_int is used to provide clk & clk2x for internal use in the FPGA
 *  
 *                 all signals with "sys_" or "ddr_" prefix include IO buffers to come 
 *                 in/out of the FPGA, don't to add IOs for them on the top level.
 *  
 *                 fpga_clk2x_out and fpga_clk_out have gone through BUFG, 
 *                 so they can be connected directly to flip/flops inside the FPGA
 *
 *       Author:  Jennnifer Tran
 *      Company:  Xilinx
 *
 *                Copyright (c) 1999 Xilinx, Inc.
 *                All rights reserved
 *
 *    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
 * 
 *   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
 *                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
 *                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
 *
 ******************************************************************************/
   
`timescale 100ps / 10ps

module clk_dlls(/*AUTOARG*/
   // Outputs
   ddr_clk_out, ddr_clk180_out, fpga_clk2x_out, fpga_clk_out, 
   dlls_locked_out, fpga_lac_clk_out, fpga_lac_clk90_out, 
   // Inputs
   sys_clk_in, sys_clk_fb_in, reset_in
   );
   input sys_clk_in, sys_clk_fb_in, reset_in;
   output ddr_clk_out, ddr_clk180_out;
   output fpga_clk2x_out, fpga_clk_out;
   output dlls_locked_out;
   output fpga_lac_clk_out, fpga_lac_clk90_out;  //Logic Accessible Clock
   
   IBUFG_SSTL2_I ibufg_sys_clk    (.I(sys_clk_in), .O(sys_clk));
   IBUFG_SSTL2_I ibufg_sys_clk_fb (.I(sys_clk_fb_in), .O(sys_clk_fb));
   
   CLKDLL dll_ext
     (.CLKIN(sys_clk), .CLKFB(sys_clk_fb), .RST(1'b0), 
      .CLK0(ddr_clk_out_o), .CLK90(), .CLK180(), .CLK270(), 
      .CLK2X(), .CLKDV(), .LOCKED(dll_ext_locked));

   //dll_int gets CLKIN from ddr_clk
   //since ddr_clk is not stable until dll_ext achieve lock, 
   //we need to hold dll_int in reset untill dll_ext_lock goes high
   wire   dll_int_reset = ~dll_ext_locked;
   CLKDLL dll_int
     (.CLKIN(sys_clk_fb), .CLKFB(fpga_clk_out), .RST(dll_int_reset), 
      .CLK0(fpga_clk_out_o), .CLK90(), .CLK180(), .CLK270(), 
      .CLK2X(fpga_clk2x_out_o), .CLKDV(), .LOCKED(dll_int_locked));
  
   OBUF_SSTL2_I obuf_ddr_clk_out (.I(ddr_clk_out_o), .O(ddr_clk_out));
   OBUF_SSTL2_I obuf_ddr_clk180_out (.I(~ddr_clk_out_o), .O(ddr_clk180_out));

   BUFG bufg_clk2x_out (.I(fpga_clk2x_out_o), .O(fpga_clk2x_out));
   BUFG bufg_clk_out (.I(fpga_clk_out_o), .O(fpga_clk_out));

   assign dlls_locked_out = dll_int_locked & dll_ext_locked;
 
   //In data_path module, we need to use clk to select b/t lower and 
   // higher bits of system data and use clk2x to generate double data output
   //However, the clk signal cannot drive non clock CLB pins in Virtex, so
   //we need to generate Logic Accessible Clock signals

   //generate Logic Accessible Clock signals
   wire #6 sys_clk_dly = sys_clk; //simulate routing delay
   FD LAC1 (.D(~sys_clk_dly), .C(fpga_clk2x_out), .Q(lac_clk1));
   FD LAC2 (.D(~lac_clk1), .C(fpga_clk2x_out), .Q(lac_clk2));
   FD LAC3 (.D(~lac_clk2), .C(fpga_clk2x_out), .Q(fpga_lac_clk_out));

   FD LAC4 (.D(sys_clk_dly), .C(~fpga_clk2x_out), .Q(lac_clk90_1));
   FD LAC5 (.D(~lac_clk90_1), .C(~fpga_clk2x_out), .Q(lac_clk90_2));
   FD LAC6 (.D(~lac_clk90_2), .C(~fpga_clk2x_out), .Q(lac_clk90_3));
   FD LAC7 (.D(~lac_clk90_3), .C(~fpga_clk2x_out), .Q(lac_clk90_4));
   FD LAC8 (.D(~lac_clk90_4), .C(~fpga_clk2x_out), .Q(fpga_lac_clk90_out));
    
endmodule // clk_dlls


