--*****************************************************************************
--
--    File Name:  clk_dlls.vhd
--      Version:  2.5
--         Date:  June 05, 2000
-- 
--  Description:   clk_ddls contain 2 dlls to deskew FPGA and DDR SDRAM clocks. 
-- 
--           dll_ext provides ddr_clk and ddr_clkb for the SDRAM and receives feedback
--                 from the SDRAM
-- 
--           dll_int is used to provide clk & clk2x for internal use in the FPGA
--  
--                 all signals with "sys_" or "ddr_" prefix include IO buffers to come 
--                 in/out of the FPGA, don't to add IOs for them on the top level.
--  
--                 fpga_clk2x_out and fpga_clk_out have gone through BUFG, 
--                 so they can be connected directly to flip/flops inside the FPGA
--
--       Author:  Jennnifer Tran
--      Company:  Xilinx
--
--                Copyright (c) 1999 Xilinx, Inc.
--                All rights reserved
--
--    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
-- 
--   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
--                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--*****************************************************************************/
library IEEE;
use IEEE.std_logic_1164.all;
library work;
use WORK.ddr_controller_comp.all;
-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on


entity clk_dlls is
  port(
    --Inputs
    sys_clk_in    : in std_logic;
    sys_clk_fb_in : in std_logic;
    reset_in      : in std_logic;
    
    --Outputs
    ddr_clk_out        : out std_logic;
    ddr_clk180_out     : out std_logic;
    fpga_clk2x_out     : out std_logic;
    fpga_clk_out       : out std_logic;
    dlls_locked_out    : out std_logic;
    fpga_lac_clk_out   : out std_logic;
    fpga_lac_clk90_out : out std_logic);
end clk_dlls;

architecture rtl of clk_dlls is
  signal sys_clk           : std_logic;
  signal sys_clk_dly       : std_logic;
  signal sys_clk_dly_inv   : std_logic;
  signal sys_clk_fb        : std_logic;
  signal ddr_clk_out_o     : std_logic;
  signal ddr_clk_out_o_inv : std_logic;
  signal ddr_clk180_out_o  : std_logic;
  signal dll_ext_locked    : std_logic;
  signal fpga_clk_outa     : std_logic;
  signal fpga_clk_out_o    : std_logic;
  signal fpga_clk2x_outa   : std_logic;
  signal fpga_clk2x_out_o  : std_logic;
  signal dll_int_locked    : std_logic;
  signal lac_clk1	         : std_logic;
  signal lac_clk2	         : std_logic;
  signal lac_clk1_inv      : std_logic;
  signal lac_clk2_inv      : std_logic;
  signal lac_clk90_1       : std_logic;
  signal lac_clk90_2       : std_logic;
  signal lac_clk90_3       : std_logic;
  signal lac_clk90_4       : std_logic;
  signal lac_clk90_1_inv   : std_logic;
  signal lac_clk90_2_inv   : std_logic;
  signal lac_clk90_3_inv   : std_logic;
  signal lac_clk90_4_inv   : std_logic;
  signal fpga_clk2x_out_inv: std_logic;
  signal dll_int_reset     : std_logic;
  signal logic0            : std_logic;

  component FD
    port(D  : in std_logic;
         C  : in std_logic;
         Q  : out std_logic);
  end component;
  component BUFG
    port(I : in std_logic;
         O : out std_logic);
  end component;
  component OBUF_SSTL2_I
    port(I : in std_logic;
         O : out std_logic);
  end component;
  component IBUFG_SSTL2_I
    port(I : in std_logic;
         O : out std_logic);
  end component;
  component IBUFG
    port(I : in std_logic;
         O : out std_logic);
  end component;
  component CLKDLL
      port ( CLKIN   : in  std_ulogic := '0';
           CLKFB   : in  std_ulogic := '0';
           RST     : in  std_ulogic := '0';
           CLK0    : out std_ulogic := '0';
           CLK90   : out std_ulogic := '0';
           CLK180  : out std_ulogic := '0';
           CLK270  : out std_ulogic := '0';
           CLK2X   : out std_ulogic := '0';
           CLKDV   : out std_ulogic := '0';
           LOCKED  : out std_ulogic := '0');
  end component;

begin

  ibufg_sys_clk:    IBUFG_SSTL2_I port map (I=>sys_clk_in,O=>sys_clk);
  ibufg_sys_clk_fb: IBUFG_SSTL2_I port map (I=>sys_clk_fb_in,O=>sys_clk_fb);

  logic0 <= '0';

  dll_ext: CLKDLL port map (CLKIN=>sys_clk,CLKFB=>sys_clk_fb, RST=>logic0,
                            CLK0=>ddr_clk_out_o,LOCKED=>dll_ext_locked);

  --dll_int gets CLKIN from ddr_clk
  --since ddr_clk is not stable until dll_ext achieve lock, 
  --we need to hold dll_int in reset untill dll_ext_lock goes high
  dll_int_reset <= not(dll_ext_locked);
  dll_int: CLKDLL port map (CLKIN=>sys_clk_fb,CLKFB=>fpga_clk_outa,RST=>dll_int_reset,
                            CLK0=>fpga_clk_out_o,CLK2X=>fpga_clk2x_out_o,
                            LOCKED=>dll_int_locked);

  ddr_clk_out_o_inv <= not(ddr_clk_out_o);

  obuf_ddr_clk_out:    OBUF_SSTL2_I port map (I=>ddr_clk_out_o,O=>ddr_clk_out);
  obuf_ddr_clk180_out: OBUF_SSTL2_I port map (I=>ddr_clk_out_o_inv,O=>ddr_clk180_out);

  bufg_clk2x_out:  BUFG port map (I=>fpga_clk2x_out_o,O=>fpga_clk2x_outa);
  bufg_clk_out:    BUFG port map (I=>fpga_clk_out_o,O=>fpga_clk_outa);

  dlls_locked_out <= dll_int_locked and dll_ext_locked;

  sys_clk_dly <= sys_clk after 1 ns;
  -- The 1ns simulate routing delay from IBUFG to
  -- first flip-flop, to prevent race
  -- condition with clk2x in functional simulation
  -- The synthesis tool can ignore this delay
  
  sys_clk_dly_inv <= not sys_clk_dly;
  lac_clk1_inv <= not lac_clk1;
  lac_clk2_inv <= not lac_clk2;
  lac_clk90_1_inv <= not lac_clk90_1;
  lac_clk90_2_inv <= not lac_clk90_2;
  lac_clk90_3_inv <= not lac_clk90_3;
  lac_clk90_4_inv <= not lac_clk90_4;
  fpga_clk2x_out_inv <= not fpga_clk2x_outa;

  fpga_clk_out <= fpga_clk_outa;
  fpga_clk2x_out <= fpga_clk2x_outa;

  --In data_path module, we need to use clk to select b/t lower and 
  --higher bits of system data and use clk2x to generate double data output
  --However, the clk signal cannot drive non clock CLB pins in Virtex, so
  --we need to generate Logic Accessible Clock signals
  
  LAC1: FD port map (D=>sys_clk_dly_inv,C=>fpga_clk2x_outa,Q=>lac_clk1);
  LAC2: FD port map (D=>lac_clk1_inv, C=>fpga_clk2x_outa,Q=>lac_clk2);
  LAC3: FD port map (D=>lac_clk2_inv, C=>fpga_clk2x_outa,Q=>fpga_lac_clk_out);
  
  LAC4: FD port map (D=>sys_clk_dly,     C=>fpga_clk2x_out_inv,Q=>lac_clk90_1);
  LAC5: FD port map (D=>lac_clk90_1_inv, C=>fpga_clk2x_out_inv,Q=>lac_clk90_2);
  LAC6: FD port map (D=>lac_clk90_2_inv, C=>fpga_clk2x_out_inv,Q=>lac_clk90_3);
  LAC7: FD port map (D=>lac_clk90_3_inv, C=>fpga_clk2x_out_inv,Q=>lac_clk90_4);
  LAC8: FD port map (D=>lac_clk90_4_inv, C=>fpga_clk2x_out_inv,Q=>fpga_lac_clk90_out);

end rtl;
