/******************************************************************************
 *
 *    File Name:  controller.v
 *      Version:  2.5
 *         Date:  June 05, 2000
 * 
 *  Description:  state machine to generate all control signals for DDR SDRAM 
 *                and other blocks of the DDR controller
 * 
 *                state machine is implemented as one-hot for performance
 *                all outputs are registered
 *                
 *                ddr_dqs_t is generated by negative edge of clk2x
 *                All other control signals are generated by positive edge of clk, 
 * 
 * 
 *       Author:  Jennnifer Tran
 *      Company:  Xilinx
 *
 *                Copyright (c) 1999 Xilinx, Inc.
 *                All rights reserved
 *
 *    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
 * 
 *   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
 *                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
 *                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
 *
 ******************************************************************************/

`timescale 100ps / 10ps

`include "../src/define.v"

module controller (/*AUTOARG*/
   // Outputs
   ddr_read_en, ddr_write_en, u_data_valid_en, ddr_dqs_t, ddr_rasb, 
   ddr_casb, ddr_web, ld_burst, ld_rcd, ld_cas_lat, u_ref_ack, 
   mrs_addr, row_addr, 
   // Inputs
   u_cmd, cntlr_reset1, cntlr_reset2, cntlr_reset3, clk, clk2x, 
   lac_clk_in, burst_end, cas_lat_end, rcd_end, burst_8, burst_2
   );   
   
   output [3:0] ddr_read_en;   
   output [7:0] ddr_write_en;
   output 	u_data_valid_en;
   output 	ddr_dqs_t;
   output 	ddr_rasb, ddr_casb, ddr_web;
   output 	ld_burst, ld_rcd, ld_cas_lat;
   output 	u_ref_ack;
   output 	mrs_addr;
   output 	row_addr;
   
   input [7:1] 	u_cmd;
   input 	cntlr_reset1, cntlr_reset2, cntlr_reset3; 
   input 	clk, clk2x, lac_clk_in, burst_end, cas_lat_end, rcd_end;
   input 	burst_8, burst_2;
   
   reg [2:0] 	ddr_cmd, next_ddr_cmd;
   reg [12:1] 	state, next_state;
   reg 		row_addr, ld_cas_lat, ld_burst;
   reg 		ld_rcd;
   reg [7:0] 	ddr_write_en;
   reg 		ddr_dqs_t_wire;
   reg 		u_ref_ack;
   reg 		mrs_addr;
   reg 		ddr_rasb_o, ddr_casb_o, ddr_web_o;
   wire 	data_read = state[`CTLR_READ_DATA];
   wire 	data_write = state[`CTLR_WRITE_DATA];
   wire 	data_read_write = data_read || data_write;
   reg 		read_cmd;
   reg 		read_cmd_reg;
   //need to register u_cmd[`SYS_READ] because it's needed to generate burst_end after the user has released the READ command
   

   /*****************/
   /* state machine */
   /*****************/
   always @ (read_cmd_reg or burst_8 or burst_2 or u_cmd or state or burst_end or cas_lat_end or rcd_end)   begin
      //has default values for outputs so synthesis tool don't infer latches
      next_state = 12'b0000_0000_0000;
      read_cmd = 0;
      casex (1'b1) 
	state[`CTLR_IDLE]   : begin
	   casex (1'b1)  
	     u_cmd[`SYS_REFRESH]:	next_state[`CTLR_REFRESH] = 1;
	     u_cmd[`SYS_PRECHARGE]: 	next_state[`CTLR_PRECHARGE] = 1;
	     u_cmd[`SYS_LOAD_MR]:       next_state[`CTLR_LOAD_MR] = 1;
	     u_cmd[`SYS_WRITE]:         next_state[`CTLR_ACT] = 1;
	     u_cmd[`SYS_READ]: begin	
		next_state[`CTLR_ACT] = 1;
		read_cmd = 1;
	     end			
	     default:   	    	next_state[`CTLR_IDLE] = 1;
	   endcase	
	end // case: state[`CTLR_IDLE]

	state[`CTLR_ACT]    : //issue ACTIVE command
	   if (rcd_end) begin
	      if (read_cmd_reg)	begin
		 next_state[`CTLR_READ] = 1;
		 read_cmd = 1;
	      end
	      else begin
		 next_state[`CTLR_WRITE] =1;
	      end
	   end
	   else begin 
	      next_state[`CTLR_ACT_WAIT] = 1;
	      if (read_cmd_reg)	read_cmd = 1;
	   end
 	   
	state[`CTLR_ACT_WAIT]    :  //issue NOP command, wait for rcd_end
	   if (rcd_end) begin
	      if (read_cmd_reg)	next_state[`CTLR_READ] = 1;
	      else next_state[`CTLR_WRITE] =1;
	   end
	   else begin
	      next_state[`CTLR_ACT_WAIT] = 1;
	      if (read_cmd_reg)	read_cmd = 1;
	   end

	state[`CTLR_READ]  : 
	  if (~cas_lat_end)	           next_state[`CTLR_READ_WAIT] = 1;
	  else	         	           next_state[`CTLR_READ_DATA] = 1;
	
	state[`CTLR_READ_WAIT]  : 
	  if (~cas_lat_end)	           next_state[`CTLR_READ_WAIT] = 1;
	  else if (burst_2)                next_state[`CTLR_IDLE] = 1;
	  else	                           next_state[`CTLR_READ_DATA] = 1;
	
	state[`CTLR_WRITE]  :	begin
	   if (burst_2)                     next_state[`CTLR_IDLE] = 1;
	   else                             next_state[`CTLR_WRITE_DATA] = 1;
	end
	
	state[`CTLR_READ_DATA]  : 
	   if (burst_8 && ~burst_end)      next_state[`CTLR_READ_DATA] = 1;
	   else                            next_state[`CTLR_IDLE] = 1;
	
	state[`CTLR_WRITE_DATA]  : begin
	   if (burst_8 && ~burst_end)      next_state[`CTLR_WRITE_DATA] = 1;
	   else  begin
	      next_state[`CTLR_IDLE] = 1;
	   end
	end

	default :               	   next_state[`CTLR_IDLE] = 1;
      endcase 
   end 
 
   /*********************************************/
   /* combinatorial next state and output logic */
   /*********************************************/  
   wire next_mrs_addr =  u_cmd[`SYS_LOAD_MR];
   wire next_row_addr =  ~(next_state[`CTLR_PRECHARGE]|| next_state[`CTLR_READ]|| next_state[`CTLR_WRITE]);
   wire next_ld_cas_lat = ~(state[`CTLR_READ] || state[`CTLR_READ_WAIT]);
   wire next_ld_burst   = ~(state[`CTLR_READ] || state[`CTLR_WRITE] || data_read_write);
   wire next_ld_rcd     = ~(state[`CTLR_ACT] || state[`CTLR_ACT_WAIT]);
   wire next_u_ref_ack      =  next_state[`CTLR_REFRESH];
   wire next_ddr_rasb_o = ~(state[`CTLR_REFRESH] || state[`CTLR_PRECHARGE] || 
       			    state[`CTLR_LOAD_MR] || state[`CTLR_ACT]);
   wire next_ddr_casb_o = ~(state[`CTLR_REFRESH] || state[`CTLR_LOAD_MR] ||
       			    state[`CTLR_READ] || state[`CTLR_WRITE]);
   wire next_ddr_web_o  = ~(state[`CTLR_PRECHARGE] || state[`CTLR_LOAD_MR] || 
	                    state[`CTLR_WRITE]);
   wire next_ddr_write_en = (state[`CTLR_WRITE] || state[`CTLR_WRITE_DATA]);
   wire next_ddr_read_en = state[`CTLR_READ_WAIT] || state[`CTLR_READ_DATA];

   reg 	next_ddr_read_en_P1, u_data_valid_en;
   always @(posedge clk)  next_ddr_read_en_P1 <= next_ddr_read_en;

   wire next_u_data_valid_en = burst_8 ? next_ddr_read_en: (next_ddr_read_en || next_ddr_read_en_P1);//for burst of 4 or 2, need to delay u_data_valid_en


   
   /******************************/
   /* output and state registers */
   /******************************/
   reg [3:0] ddr_read_en;
   reg 	     ddr_rasb_o_P1, ddr_casb_o_P1, ddr_web_o_P1;
   //ddr control signals are generated on negative edge clock to guarantee DDR hold time
   always @ (negedge clk or posedge cntlr_reset1) begin
      if (cntlr_reset1) begin
	 ddr_rasb_o <= 1;
	 ddr_casb_o <= 1;
	 ddr_web_o <=1;	 
      end
      else begin
	 ddr_rasb_o <= ddr_rasb_o_P1;
	 ddr_casb_o <= ddr_casb_o_P1;
	 ddr_web_o <= ddr_web_o_P1;
      end
   end // always @ (posedge clk or posedge cntlr_reset1)
   
   always @ (posedge clk) begin
      ddr_rasb_o_P1 <=  next_ddr_rasb_o;
      ddr_casb_o_P1 <=  next_ddr_casb_o;
      ddr_web_o_P1  <=  next_ddr_web_o;
   end

   always @ (posedge clk or posedge cntlr_reset2) begin
      if (cntlr_reset2) begin
	 state[`CTLR_IDLE] <= 1;
	 state[`CTLR_REFRESH] <= 0;
	 state[`CTLR_PRECHARGE] <= 0;
	 state[`CTLR_LOAD_MR] <= 0;
	 state[`CTLR_ACT] <= 0;
	 state[`CTLR_ACT_WAIT] <= 0;
	 state[`CTLR_READ] <= 0;
	 state[`CTLR_WRITE] <= 0;
	 state[`CTLR_READ_WAIT] <= 0;
	 state[`CTLR_READ_DATA] <= 0;
	 state[`CTLR_WRITE_DATA] <= 0;
	 read_cmd_reg <= 0;
      end
      else begin
	 read_cmd_reg <= read_cmd;
	 ddr_read_en <= {4{next_ddr_read_en_P1}};
	 mrs_addr <= next_mrs_addr;
	 row_addr <= next_row_addr;
	 ld_cas_lat <= next_ld_cas_lat;
	 ld_burst <= next_ld_burst;
	 ld_rcd <= next_ld_rcd;
	 u_ref_ack <= next_u_ref_ack;
	 state <= next_state;
	 u_data_valid_en <= next_u_data_valid_en;
      end
   end // always @ (posedge clk)


   //ddr_write_en is used to enable ddr_dq signals
   reg next_ddr_write_en_P1;
   always @(posedge clk) 
     next_ddr_write_en_P1 <= next_ddr_write_en;
   
   always @(posedge clk2x)
      ddr_write_en <= {8{next_ddr_write_en_P1}};



   //ddr_dqs_t is used to generate ddr_dqs signals
   reg ddr_dqs_t_tmp1, ddr_dqs_t_tmp2;
   always @(posedge clk) begin
      ddr_dqs_t_tmp1 <= next_ddr_write_en;
      ddr_dqs_t_tmp2 <= ddr_dqs_t_tmp1;
   end
   
   //create local lac_clk and also generate lac_clk_out
   FD LAC1 (.D(~lac_clk_in), .C(~clk2x), .Q(lac_clk_out));
   FD LAC2 (.D(~lac_clk_in), .C(~clk2x), .Q(lac_clk_wire));
   wire #15 lac_clk = lac_clk_wire;

   reg 	    ddr_dqs_t;
   always @ (negedge clk2x or posedge cntlr_reset3)
      if (cntlr_reset3)	ddr_dqs_t <= 1;
      else if (lac_clk) ddr_dqs_t <= ~(ddr_dqs_t_tmp1 || next_ddr_write_en);


   /********************************************/
   /* IO buffers for DDR SDRAM control signals */
   /********************************************/
   OBUF_SSTL2_I O_ddr_ras (.O(ddr_rasb), .I(ddr_rasb_o));
   OBUF_SSTL2_I O_ddr_cas (.O(ddr_casb), .I(ddr_casb_o));
   OBUF_SSTL2_I O_ddr_we  (.O(ddr_web), .I(ddr_web_o));
endmodule // controller
