/******************************************************************************
 *
 *    File Name:  data_path.v
 *      Version:  2.5
 *         Date:  June 05, 2000
 * 
 *  Description:  generates ddr_dq and ddr_dqs during write cycles.  
 *                Receives ddr_dq during read cycle and generate u_data_o and u_data_valid
 * 
 *                Write:
 *                32-bit u_data_i is muxed into 16-bit ddr_dq with clk2x. 
 *                lac_clk selects between higher and lower data.
 *                tristate signal for ddr_dq (ddr_dq_t) is generated from ddr_write_en
 *                ddr_dqs is generated by a toggle flipflop and clocked by negedge of clk2x.
 *                this ensures dqs edges are at the center of dq
 * 
 *                Read:
 *                16-bit ddr_dq is clocked in by clk2x.  They are then assembled into
 *                32-bit data with higher bits registered by posedge of clk, and lower bits
 *                registered by negedge clk.
 * 
 *                All inputs and outputs to DDR SDRAM are registered in the IOB to guarantee
 *                minimum clock-to-out delay 
 *
 * 
 *       Author:  Jennnifer Tran
 *      Company:  Xilinx
 *
 *                Copyright (c) 1999 Xilinx, Inc.
 *                All rights reserved
 *
 *    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
 * 
 *   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
 *                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
 *                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
 *                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
 *
 ******************************************************************************/

`timescale  100 ps / 10 ps
`include "../src/define.v"

module data_path (/*AUTOARG*/
   // Outputs
   u_data_o, ddr_dqs, u_data_valid, 
   // Inouts
   ddr_dq, 
   // Inputs
   u_data_i, ddr_write_en, ddr_dqs_t, ddr_read_en, u_data_valid_en, 
   clk2x, clk, lac_clk_in, cas_lat_half, burst_8
   );

   input [`U_DATA_MSB:0]     u_data_i;
   input [7:0] 		     ddr_write_en;
   input 		     ddr_dqs_t;
   input [`ENABLE_MSB:0]     ddr_read_en;
   input 		     u_data_valid_en;
   input 		     clk2x, clk, lac_clk_in;	
   input [1:0]		     cas_lat_half;
   input 		     burst_8;
   inout [`DDR_DATA_MSB:0]   ddr_dq;
   output [`U_DATA_MSB:0]    u_data_o;
   output [1:0] 	     ddr_dqs;
   output 		     u_data_valid;

   /**************************************
    * pipeline/duplicate control signals *
    **************************************/
   //pipeline ddr_write_en
   reg [7:0] 		     ddr_write_en_P1_5;
   always @(posedge clk) begin
	ddr_write_en_P1_5 <= ddr_write_en;
   end

   //duplicate tristate signal to reduce net delay
   wire [3:0] ddr_dq_t_wire;
   assign     ddr_dq_t_wire[0] = ~(ddr_write_en_P1_5[0]);
   assign     ddr_dq_t_wire[1] = ~(ddr_write_en_P1_5[1]);
   assign     ddr_dq_t_wire[2] = ~(ddr_write_en_P1_5[2]);
   assign     ddr_dq_t_wire[3] = ~(ddr_write_en_P1_5[3]);
   
   //clk is used to select b/t lower and higher bits of write_data
   //and use clk2x to generate double data output
   //However, the clk signal cannot drive non clock inputs in Virtex, so
   //we need to use Logic Accessible Clock signal

   //pipeline Logic Accessible Clock signal to reduce delay
   FD LAC_temp1 (.D(~lac_clk_in), .C(clk2x), .Q(lac_clk_q1));
   FD LAC_temp2 (.D(~lac_clk_in), .C(clk2x), .Q(lac_clk_q2));
   wire #(15) lac_clk_tmp1 = lac_clk_q1; //simulate clk-to-out delay
   wire #(15) lac_clk_tmp2 = lac_clk_q2; //simulate clk-to-out delay
   
   FD LAC1 (.D(~lac_clk_tmp1), .C(clk2x), .Q(lac_clkq1));
   FD LAC2 (.D(~lac_clk_tmp1), .C(clk2x), .Q(lac_clkq2));
   FD LAC3 (.D(~lac_clk_tmp2), .C(clk2x), .Q(lac_clkq3));
   FD LAC4 (.D(~lac_clk_tmp2), .C(clk2x), .Q(lac_clkq4));
   
   wire [4:1] lac_clk;
   assign #(15) lac_clk[1] = lac_clkq1;
   assign #(15) lac_clk[2] = lac_clkq2;
   assign #(15) lac_clk[3] = lac_clkq3;
   assign #(15) lac_clk[4] = lac_clkq4;
 
    
   /***************
    * Write cycle *
    ***************/

   //generate ddr_dq
   reg [`U_DATA_MSB:0] write_data;

   //delay input data for 5 clk cycles
   reg [`U_DATA_MSB:0]  u_data_i_P1, u_data_i_P2, u_data_i_P3, u_data_i_P4, u_data_i_P5;
   always @(posedge clk) begin
      u_data_i_P1 <= u_data_i;
      u_data_i_P2 <= u_data_i_P1;
      u_data_i_P3 <= u_data_i_P2;
      u_data_i_P4 <= u_data_i_P3;
      u_data_i_P5 <= u_data_i_P4;
   end
   always @(negedge clk)
      write_data <= u_data_i_P5;

         
   //mux 32-bit write data into 16-bit ddr_dq   
   reg [`DDR_DATA_MSB:0]     ddr_dq_o, ddr_dq_o_tmp;
   wire [`DDR_DATA_MSB:0]     ddr_dq_t;
   always @(posedge clk2x) begin
      //select lower bits
      if (ddr_write_en[0] && lac_clk[1])
	ddr_dq_o_tmp[1:0]  <= write_data[1:0];
      if (ddr_write_en[1] && lac_clk[1])
	ddr_dq_o_tmp[3:2]  <= write_data[3:2];
      
      if (ddr_write_en[2] && lac_clk[2])
	ddr_dq_o_tmp[5:4]  <= write_data[5:4];
      if (ddr_write_en[3] && lac_clk[2])
	ddr_dq_o_tmp[7:6]  <= write_data[7:6];
      
      if (ddr_write_en[4] && lac_clk[3])
	ddr_dq_o_tmp[9:8]  <= write_data[9:8];
      if (ddr_write_en[5] && lac_clk[3])
	ddr_dq_o_tmp[11:10]  <= write_data[11:10];
      
      if (ddr_write_en[6] && lac_clk[4])
	ddr_dq_o_tmp[13:12]  <= write_data[13:12];
      if (ddr_write_en[7] && lac_clk[4])
	ddr_dq_o_tmp[15:14]  <= write_data[15:14];

      //select higher bits
      if (ddr_write_en[0] && ~lac_clk[1])
	ddr_dq_o_tmp[1:0]  <= write_data[17:16];
      if (ddr_write_en[1] && ~lac_clk[1])
	ddr_dq_o_tmp[3:2]  <= write_data[19:18];
      if (ddr_write_en[2] && ~lac_clk[2])
	ddr_dq_o_tmp[5:4]  <= write_data[21:20];
      if (ddr_write_en[3] && ~lac_clk[2])
	ddr_dq_o_tmp[7:6]  <= write_data[23:22];
      if (ddr_write_en[4] && ~lac_clk[3])
	ddr_dq_o_tmp[9:8]  <= write_data[25:24];
      if (ddr_write_en[5] && ~lac_clk[3])
	ddr_dq_o_tmp[11:10]  <= write_data[27:26];
      if (ddr_write_en[6] && ~lac_clk[4])
	ddr_dq_o_tmp[13:12]  <= write_data[29:28];
      if (ddr_write_en[7] && ~lac_clk[4])
	ddr_dq_o_tmp[15:14]  <= write_data[31:30];
   end // always @ (posedge clk2x)

   reg [15:0] ddr_dq_o_tmp2;   
   always @(posedge clk2x) begin
      ddr_dq_o <= #18 ddr_dq_o_tmp;  //simulate clk-to-out delay on IOB flip-flop
   end
   

   //generate tristate signal for ddr_dq
   //ddr_dq_t f/f should be in IOB
   //always @(posedge clk2x) begin
   //	ddr_dq_t[3:0]   <= #18 {4{ddr_dq_t_wire[0]}};
   //	ddr_dq_t[7:4]   <= #18 {4{ddr_dq_t_wire[1]}};
   //	ddr_dq_t[11:8]  <= #18 {4{ddr_dq_t_wire[2]}};
   //	ddr_dq_t[15:12] <= #18 {4{ddr_dq_t_wire[3]}};
   //end 
   //There was a bug in D.19 place&route where the tristate f/f can't
   //be merged into the IOB.  To get around the bug, we need to
   //instantiate FDs directly
   wire [15:0] ddr_dq_t_reg;
   assign  #18    ddr_dq_t  =  ddr_dq_t_reg;
   FD  u_ddr_dq_t0  (.D(ddr_dq_t_wire[0]), .C(clk2x), .Q(ddr_dq_t_reg[0]));
   FD  u_ddr_dq_t1  (.D(ddr_dq_t_wire[0]), .C(clk2x), .Q(ddr_dq_t_reg[1]));
   FD  u_ddr_dq_t2  (.D(ddr_dq_t_wire[0]), .C(clk2x), .Q(ddr_dq_t_reg[2]));
   FD  u_ddr_dq_t3  (.D(ddr_dq_t_wire[0]), .C(clk2x), .Q(ddr_dq_t_reg[3]));
      	  							 
   FD  u_ddr_dq_t4  (.D(ddr_dq_t_wire[1]), .C(clk2x), .Q(ddr_dq_t_reg[4]));
   FD  u_ddr_dq_t5  (.D(ddr_dq_t_wire[1]), .C(clk2x), .Q(ddr_dq_t_reg[5]));
   FD  u_ddr_dq_t6  (.D(ddr_dq_t_wire[1]), .C(clk2x), .Q(ddr_dq_t_reg[6]));
   FD  u_ddr_dq_t7  (.D(ddr_dq_t_wire[1]), .C(clk2x), .Q(ddr_dq_t_reg[7]));
      	  							 
   FD  u_ddr_dq_t8  (.D(ddr_dq_t_wire[2]), .C(clk2x), .Q(ddr_dq_t_reg[8]));
   FD  u_ddr_dq_t9  (.D(ddr_dq_t_wire[2]), .C(clk2x), .Q(ddr_dq_t_reg[9]));
   FD  u_ddr_dq_t10  (.D(ddr_dq_t_wire[2]), .C(clk2x), .Q(ddr_dq_t_reg[10]));
   FD  u_ddr_dq_t11  (.D(ddr_dq_t_wire[2]), .C(clk2x), .Q(ddr_dq_t_reg[11]));
      	   							  
   FD  u_ddr_dq_t12  (.D(ddr_dq_t_wire[3]), .C(clk2x), .Q(ddr_dq_t_reg[12]));
   FD  u_ddr_dq_t13  (.D(ddr_dq_t_wire[3]), .C(clk2x), .Q(ddr_dq_t_reg[13]));
   FD  u_ddr_dq_t14  (.D(ddr_dq_t_wire[3]), .C(clk2x), .Q(ddr_dq_t_reg[14]));
   FD  u_ddr_dq_t15  (.D(ddr_dq_t_wire[3]), .C(clk2x), .Q(ddr_dq_t_reg[15]));

   //generate dqs
   //toggle f/f will be implemented in CLB
   reg [1:0] ddr_dqs_o_tmp;
   reg 	     ddr_dqs_tog1, ddr_dqs_tog2;
   always @(negedge clk2x) begin
      ddr_dqs_tog1 <= ddr_dqs_t;
      ddr_dqs_tog2 <= ddr_dqs_tog1;
   end
   always @(negedge clk2x) begin
      if (ddr_dqs_t)
	ddr_dqs_o_tmp <= #1 2'b00;
      else 
	ddr_dqs_o_tmp <= #1 ~ddr_dqs_o_tmp;
   end
   reg [1:0] ddr_dqs_o_tmp_P1, ddr_dqs_o_tmp_P2;
   //delay ddr_dqs_o_tmp by 2 clk2x cycles
   always @(negedge clk2x) begin
      ddr_dqs_o_tmp_P1 <= ddr_dqs_o_tmp;
      ddr_dqs_o_tmp_P2 <= ddr_dqs_o_tmp_P1;
   end
   
   //the following f/fs will be in IOB
   reg [1:0] ddr_dqs_o_reg, ddr_dqs_t_P1, ddr_dqs_t_reg;
   always @(negedge clk2x)  begin
      ddr_dqs_o_reg <= #24 ddr_dqs_o_tmp_P2;  //simulate clk-to-out delay on IOB flip-flop
      ddr_dqs_t_P1 <= {2{ddr_dqs_t}};
      ddr_dqs_t_reg <= #24 ddr_dqs_t_P1;   //simulate clk-to-out delay on IOB flip-flop
   end
     
   OBUFT_SSTL2_II ddr_dqs0 (.I(ddr_dqs_o_reg[0]), .O(ddr_dqs[0]), .T(ddr_dqs_t_reg[0]));
   OBUFT_SSTL2_II ddr_dqs1 (.I(ddr_dqs_o_reg[1]), .O(ddr_dqs[1]), .T(ddr_dqs_t_reg[1]));


   /**************
    * READ cycle *
    **************/

   //pipeline read control signals
   reg u_data_valid_en_P1, u_data_valid_en_P2, u_data_valid_en_P3, u_data_valid_en_P4;
   always @(posedge clk) begin
      u_data_valid_en_P1 <= u_data_valid_en;
      u_data_valid_en_P2 <= u_data_valid_en_P1;
      u_data_valid_en_P3 <= u_data_valid_en_P2;
      u_data_valid_en_P4 <= u_data_valid_en_P3;
   end
 
   // read in data
   wire [`DDR_DATA_MSB:0]    ddr_dq_i;
   reg [`DDR_DATA_MSB:0]     read_data;
   always @(posedge clk2x) begin
     //register 16-bit data from ddr sdram
      if (ddr_read_en[0]) read_data[3:0] <= ddr_dq_i[3:0];
      if (ddr_read_en[1]) read_data[7:4] <= ddr_dq_i[7:4];
      if (ddr_read_en[2]) read_data[11:8] <= ddr_dq_i[11:8];
      if (ddr_read_en[3]) read_data[15:12] <= ddr_dq_i[15:12];
   end

   //split 16-bit DDR read_data into 32-bit u_ad
   wire [`DDR_DATA_MSB:0] #1 read_data_dly = read_data; //simulate clock-to-out delay
   reg  [`U_DATA_MSB:0]  u_data_o_20, u_data_o_20_tmp, u_data_o_20_P1;
   reg  [`U_DATA_MSB:0]  u_data_o_25, u_data_o_25_tmp, u_data_o_25_P1;
   wire [`U_DATA_MSB:0] u_data_o;

   //CAS latency 2.5
   always @(negedge clk) u_data_o_25_tmp[`U_DATA_MSB:`DDR_DATA_MSB+1] <= read_data_dly;
   always @(posedge clk) u_data_o_25_tmp[`DDR_DATA_MSB:0]  <= read_data_dly;
   always @(negedge clk) u_data_o_25_P1 <= u_data_o_25_tmp;
   always @(posedge clk) u_data_o_25 <= u_data_o_25_P1;

   // CAS latency 2
   always @(posedge clk) u_data_o_20_tmp[`U_DATA_MSB:`DDR_DATA_MSB+1]  <= read_data_dly;
   always @(negedge clk) u_data_o_20_tmp[`DDR_DATA_MSB:0]  <= read_data_dly;
   always @(posedge clk) u_data_o_20 <= u_data_o_20_tmp;

   assign u_data_o = cas_lat_half[0]? u_data_o_25 : u_data_o_20;
	  
  //generate u_data_valid   
   reg u_data_valid_20, u_data_valid_25;
   wire u_data_valid;
   always @(posedge clk)
     if (u_data_valid_en_P4 && u_data_valid_en_P1) u_data_valid_25 <= 1;
     else       u_data_valid_25 <= 0;
   
    always @(posedge clk)
      if (u_data_valid_en_P3 && u_data_valid_en) u_data_valid_20 <= 1;
      else       u_data_valid_20 <= 0;

   assign u_data_valid = cas_lat_half[1] ? u_data_valid_25 : u_data_valid_20;
  
   //IO buffers to interface with DDR SDRAM *
   wire [15:0] ddr_dq_i_wire;
   assign    ddr_dq_i = ddr_dq_i_wire;
   
   ddr_dq_io_16 ddr_dq_io (.ddr_dq_i(ddr_dq_i_wire), .ddr_dq(ddr_dq), .ddr_dq_o(ddr_dq_o), .ddr_dq_t(ddr_dq_t));
   
endmodule 




