--*****************************************************************************
--
--    File Name:  data_path.vhd
--      Version:  2.5
--         Date:  June 05, 2000
-- 
--  Description:  generates ddr_dq and ddr_dqs during write cycles.  
--                Receives ddr_dq during read cycle and generate u_data_o and u_data_valid
-- 
--                Write:
--                32-bit u_data_i is muxed into 16-bit ddr_dq with clk2x. 
--                lac_clk selects between higher and lower data.
--                tristate signal for ddr_dq (ddr_dq_t) is generated from ddr_write_en
--                ddr_dqs is generated by a toggle flipflop and clocked by negedge of clk2x.
--                this ensures dqs edges are at the center of dq
-- 
--                Read:
--                16-bit ddr_dq is clocked in by clk2x.  They are then assembled into
--                32-bit data with higher bits registered by posedge of clk, and lower bits
--                registered by negedge clk.
-- 
--                All inputs and outputs to DDR SDRAM are registered in the IOB to guarantee
--                minimum clock-to-out delay 
--
-- 
--       Author:  Jennnifer Tran
--      Company:  Xilinx
--
--                Copyright (c) 1999 Xilinx, Inc.
--                All rights reserved
--
--    DDR SDRAM:  64M (1M x 16Bit x 4Bank)
-- 
--   Disclaimer:  THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY 
--                WHATSOEVER AND XILINX SPECIFICALLY DISCLAIMS ANY 
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--*****************************************************************************/

library IEEE;
use IEEE.std_logic_1164.all;
library work;
use WORK.ddr_controller_comp.all;
-- synopsys translate_off
library unisim;
use unisim.vcomponents.all;
-- synopsys translate_on

entity data_path is
  port(
    --Outputs
    u_data_o	    : out std_logic_vector(U_DATA_MSB downto 0);
    ddr_dqs	    : out std_logic_vector(1 downto 0);
    u_data_valid    : out std_logic;
    
    --Inout
    ddr_dq	    : inout std_logic_vector(DDR_DATA_MSB downto 0);
    
    --Inputs
    u_data_i	    : in std_logic_vector(U_DATA_MSB downto 0);
    ddr_write_en    : in std_logic_vector(7 downto 0);
    ddr_dqs_t	    : in std_logic;
    u_data_valid_en : in std_logic;
    cas_lat_half    : in std_logic_vector(1 downto 0);
    burst_8         : in std_logic;        
    ddr_read_en     : in std_logic_vector(ENABLE_MSB downto 0);
    clk2x	    : in std_logic;
    clk             : in std_logic;
    lac_clk_in	    : in std_logic);
end data_path;

architecture rtl of data_path is
  signal u_data_valid_en_p1     : std_logic;
  signal u_data_valid_en_p2     : std_logic;
  signal u_data_valid_en_p3     : std_logic;
  signal u_data_valid_en_p4     : std_logic;
  signal u_data_o_25_tmp        : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_o_25_p1         : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_o_25            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_o_20_tmp        : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_o_20_p1         : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_o_20            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_valid_25        : std_logic;
  signal u_data_valid_20        : std_logic;  
  signal ddr_dqs_tog1           : std_logic;
  signal ddr_dqs_tog2           : std_logic;
  signal u_data_i_p1            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_i_p2            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_i_p3            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_i_p4            : std_logic_vector(U_DATA_MSB downto 0);
  signal u_data_i_p5            : std_logic_vector(U_DATA_MSB downto 0);
  signal ddr_write_en_p1_5      : std_logic_vector(7 downto 0);
  
  signal ddr_dq_t_wire          : std_logic_vector(3 downto 0);
  signal lac_clk                : std_logic_vector(4 downto 1);
  signal lac_clk_q1             : std_logic;
  signal lac_clk_q2             : std_logic;
  signal lac_clk_inv		: std_logic;
  signal lac_clk_tmp1_inv 	: std_logic;
  signal lac_clk_tmp2_inv 	: std_logic;
  signal lac_clkq1        	: std_logic;
  signal lac_clkq2        	: std_logic;  
  signal lac_clkq3        	: std_logic;
  signal lac_clkq4        	: std_logic;  

  signal write_data		: std_logic_vector(31 downto 0);
  signal ddr_dq_t		: std_logic_vector(15 downto 0);
  signal ddr_dq_o		: std_logic_vector(15 downto 0);
  signal ddr_dq_i		: std_logic_vector(15 downto 0);
  signal ddr_dq_o_tmp           : std_logic_vector(15 downto 0);
  signal ddr_dqs_o_tmp          : std_logic_vector(1 downto 0);
  signal ddr_dqs_o_tmp_n        : std_logic_vector(1 downto 0);
  signal ddr_dqs_o_tmp_p1       : std_logic_vector(1 downto 0);
  signal ddr_dqs_o_tmp_p2       : std_logic_vector(1 downto 0);
  signal ddr_dqs_o_reg          : std_logic_vector(1 downto 0);
  signal ddr_dqs_t_p1           : std_logic_vector(1 downto 0);
  signal ddr_dqs_t_reg          : std_logic_vector(1 downto 0);
  signal read_data		: std_logic_vector(15 downto 0);
  signal read_data_dly		: std_logic_vector(15 downto 0);
  signal ddr_dq_tmp		: std_logic_vector(15 downto 0);

  component FD
    port(D  : in std_logic;
         C  : in std_logic;
         Q  : out std_logic);
  end component;
  component OBUFT_SSTL2_II
    port(I : in std_logic;
         T : in std_logic;
         O : out std_logic);
  end component;
  component IOBUF_SSTL2_II
    port(I : in std_logic;
         IO: inout std_logic;
         T : in std_logic;
         O : out std_logic);
  end component;
 
begin
-------------------------------------------------------------------------------
-- pipeline/duplicate control signals------------------------------------------
-------------------------------------------------------------------------------
  --pipeline ddr_write_en
  process (clk)
  begin
    if (clk 'event and clk = '1') then
      ddr_write_en_p1_5 <=ddr_write_en;
    end if;
  end process;
  
  --duplicate tristate signal to reduce net delay
  g0: for i in 0 to 3 generate
    ddr_dq_t_wire(i)<= not ddr_write_en_p1_5(i);
  end generate;
  
  --clk is used to select b/t lower and higher bits of write_data
  --and use clk2x to generate double data output
  --However, the clk signal cannot drive non clock inputs in FPGAs, so
  --we need to use Logic Accessible Clock signal
  
  --pipeline Logic Accessible Clock signal to reduce delay
  lac_clk_inv <= not (lac_clk_in);
  
  LAC_tmp1: FD port map(D=>lac_clk_inv,C=>clk2x,Q=>lac_clk_q1);
  LAC_tmp2: FD port map(D=>lac_clk_inv,C=>clk2x,Q=>lac_clk_q2);
  lac_clk_tmp1_inv <= not (lac_clk_q1) after 1 ns;  -- simulate clk-to-out delay
  lac_clk_tmp2_inv <= not (lac_clk_q2) after 1 ns;  -- simulate clk-to-out delay
      
  LAC1:   FD port map(D=>lac_clk_tmp1_inv,C=>clk2x,Q=>lac_clkq1);
  LAC2:   FD port map(D=>lac_clk_tmp1_inv,C=>clk2x,Q=>lac_clkq2);
  LAC3:   FD port map(D=>lac_clk_tmp2_inv,C=>clk2x,Q=>lac_clkq3);
  LAC4:   FD port map(D=>lac_clk_tmp2_inv,C=>clk2x,Q=>lac_clkq4);
  lac_clk(1) <= lac_clkq1 after 1 ns;
  lac_clk(2) <= lac_clkq2 after 1 ns;
  lac_clk(3) <= lac_clkq3 after 1 ns;
  lac_clk(4) <= lac_clkq4 after 1 ns;

-------------------------------------------------------------------------------
-- Write cycle ----------------------------------------------------------------
-------------------------------------------------------------------------------

  --generate ddr_dq
  process(clk)
  begin
    if (clk 'event and clk = '1') then
      u_data_i_p1 <= u_data_i;
      u_data_i_p2 <= u_data_i_p1;
      u_data_i_p3 <= u_data_i_p2;
      u_data_i_p4 <= u_data_i_p3;
      u_data_i_p5 <= u_data_i_p4;
    end if;
  end process;
  
  process(clk)
  begin
    if (clk 'event and clk = '0') then
      write_data <= u_data_i_P5;
    end if;
  end process;
  
  --mux 32-bit write data into 16-bit ddr_dq
  process(clk2x) begin
    if (clk2x 'event and clk2x = '1') then
      --select lower bits
      if (ddr_write_en(0)= '1' and lac_clk(1) = '1') then
        ddr_dq_o_tmp(1 downto 0) <= write_data(1 downto 0);
      end if;
      if (ddr_write_en(1)= '1' and lac_clk(1) = '1') then
        ddr_dq_o_tmp(3 downto 2) <= write_data(3 downto 2);
      end if;
      if (ddr_write_en(2)= '1' and lac_clk(2) = '1') then
        ddr_dq_o_tmp(5 downto 4) <= write_data(5 downto 4);
      end if;
      if (ddr_write_en(3)= '1' and lac_clk(2) = '1') then
        ddr_dq_o_tmp(7 downto 6) <= write_data(7 downto 6);
      end if;
      if (ddr_write_en(4)= '1' and lac_clk(3) = '1') then
        ddr_dq_o_tmp(9 downto 8) <= write_data(9 downto 8);
      end if;
      if (ddr_write_en(5)= '1' and lac_clk(3) = '1') then
        ddr_dq_o_tmp(11 downto 10) <= write_data(11 downto 10);
      end if;
      if (ddr_write_en(6)= '1' and lac_clk(4) = '1') then
        ddr_dq_o_tmp(13 downto 12) <= write_data(13 downto 12);
      end if;
      if (ddr_write_en(7)= '1' and lac_clk(4) = '1') then
        ddr_dq_o_tmp(15 downto 14) <= write_data(15 downto 14);
      end if;
      --select higher bits
      if (ddr_write_en(0)= '1' and lac_clk(1) = '0') then
        ddr_dq_o_tmp(1 downto 0) <= write_data(17 downto 16);
      end if;
      if (ddr_write_en(1)= '1' and lac_clk(1) = '0') then
        ddr_dq_o_tmp(3 downto 2) <= write_data(19 downto 18);
      end if;
      if (ddr_write_en(2)= '1' and lac_clk(2) = '0') then
        ddr_dq_o_tmp(5 downto 4) <= write_data(21 downto 20);
      end if;
      if (ddr_write_en(3)= '1' and lac_clk(2) = '0') then
        ddr_dq_o_tmp(7 downto 6) <= write_data(23 downto 22);
      end if;
      if (ddr_write_en(4)= '1' and lac_clk(3) = '0') then
        ddr_dq_o_tmp(9 downto 8) <= write_data(25 downto 24);
      end if;
      if (ddr_write_en(5)= '1' and lac_clk(3) = '0') then
        ddr_dq_o_tmp(11 downto 10) <= write_data(27 downto 26);
      end if;
      if (ddr_write_en(6)= '1' and lac_clk(4) = '0') then
        ddr_dq_o_tmp(13 downto 12) <= write_data(29 downto 28);
      end if;
      if (ddr_write_en(7)= '1' and lac_clk(4) = '0') then
        ddr_dq_o_tmp(15 downto 14) <= write_data(31 downto 30);
      end if;
    --additional reg to be pulled into IOB   
    ddr_dq_o <= ddr_dq_o_tmp after 1.8 ns;  --simulate clk-to-out delay
    end if;
  end process;
  
  --generate tristate signal for ddr_dq
  -- ddr_dq_t f/f should be pulled into IOB
  process(clk2x) begin
     if (clk2x 'event and clk2x = '1') then
       g3: for i in 0 to 3 loop
          ddr_dq_t(i) <= ddr_dq_t_wire(0) after 2 ns;  --simulate clk-out delay
       end loop;
       g4: for i in 4 to 7 loop
          ddr_dq_t(i) <= ddr_dq_t_wire(1) after 2 ns;  --simulate clk-out delay
       end loop;
       g5: for i in 8 to 11 loop
          ddr_dq_t(i) <= ddr_dq_t_wire(2) after 2 ns;  --simulate clk-out delay
       end loop;
       g6: for i in 12 to 15 loop
          ddr_dq_t(i) <= ddr_dq_t_wire(3) after 2 ns;  --simulate clk-out delay
       end loop;
     end if;
  end process;
  
  
  --generate dqs
  --toggle f/f will be implemented in CLB
  process(clk2x)
  begin
    if (clk2x 'event and clk2x = '0') then
      ddr_dqs_tog1 <= ddr_dqs_t;
      ddr_dqs_tog2 <= ddr_dqs_tog1;
    end if;
  end process;
  
  ddr_dqs_o_tmp_n <= not(ddr_dqs_o_tmp);
  process(clk2x)
  begin
    if (clk2x 'event and clk2x = '0') then
      if (ddr_dqs_t = '1') then
        ddr_dqs_o_tmp <= "00" after 1 ns;
      else
        ddr_dqs_o_tmp <= ddr_dqs_o_tmp_n after 1 ns;
      end if;
    end if;
  end process;
  -- delay ddr_dqs_o_tmp by 2 clk2x cycles
  process(clk2x)
  begin
    if (clk2x 'event and clk2x = '0') then
      ddr_dqs_o_tmp_p1 <= ddr_dqs_o_tmp;
      ddr_dqs_o_tmp_p2 <= ddr_dqs_o_tmp_p1;
    end if;
  end process;
  
  -- the following f/fs will be in IOB
  process(clk2x) begin
     if (clk2x 'event and clk2x = '0') then
        ddr_dqs_o_reg <= ddr_dqs_o_tmp_p2 after 2 ns;
        ddr_dqs_t_p1  <= ddr_dqs_t & ddr_dqs_t;
        ddr_dqs_t_reg <= ddr_dqs_t_p1 after 2 ns;
     end if;
  end process;
  
  ddr_dqs0: OBUFT_SSTL2_II port map(I=>ddr_dqs_o_reg(0),O=>ddr_dqs(0),T=>ddr_dqs_t_reg(0));
  ddr_dqs1: OBUFT_SSTL2_II port map(I=>ddr_dqs_o_reg(1),O=>ddr_dqs(1),T=>ddr_dqs_t_reg(1));

-------------------------------------------------------------------------------
-- READ cycle------------------------------------------------------------------
-------------------------------------------------------------------------------     
  --pipeline read control signals
  process(clk) begin
     if (clk 'event and clk = '1') then
        u_data_valid_en_p1 <= u_data_valid_en;
        u_data_valid_en_p2 <= u_data_valid_en_p1;
        u_data_valid_en_p3 <= u_data_valid_en_p2;
        u_data_valid_en_p4 <= u_data_valid_en_p3;
     end if;
  end process;
  

  --read in data   
  --register 16-bit data from ddr sdram
  process(clk2x) begin
     if (clk2x 'event and clk2x = '1') then
        g4: for i in 0 to ENABLE_MSB loop
          if (ddr_read_en(i) = '1') then
             read_data(3+(4*i) downto (4*i))<=ddr_dq_i(3+(4*i) downto (4*i));
          end if;
        end loop;
     end if;
  end process;
  
  --join 16-bit DDR read_data into 32-bit u_data
  read_data_dly <= read_data after 1 ns; --simulate clock-to-out delay

  -- CAS latency 2.5
  process(clk) begin
    if (clk 'event and clk = '0') then
       u_data_o_25_tmp(U_DATA_MSB downto DDR_DATA_MSB + 1) <= read_data_dly;
       u_data_o_25_p1 <= u_data_o_25_tmp;
    end if;
  end process; 
  process(clk) begin
    if (clk 'event and clk = '1') then
       u_data_o_25_tmp(DDR_DATA_MSB downto 0) <= read_data_dly;
       u_data_o_25 <= u_data_o_25_p1;
    end if;
  end process; 
  -- CAS latency 2
  process(clk) begin
    if (clk 'event and clk = '1') then
       u_data_o_20_tmp(U_DATA_MSB downto DDR_DATA_MSB + 1) <= read_data_dly;
       u_data_o_20 <= u_data_o_20_tmp;
    end if;
  end process; 
  process(clk) begin
    if (clk 'event and clk = '0') then
       u_data_o_20_tmp(DDR_DATA_MSB downto 0) <= read_data_dly;
    end if;
  end process; 

  process (cas_lat_half, u_data_o_25, u_data_o_20)
  begin  -- process
    if (cas_lat_half(0) = '1') then
      u_data_o <= u_data_o_25;
    else
      u_data_o <= u_data_o_20;
    end if;
  end process;
  
  --generate u_data_valid
  process(clk) begin
    if (clk 'event and clk = '1') then
       if(u_data_valid_en_p4 = '1' and u_data_valid_en_p1 = '1') then 
          u_data_valid_25 <= '1';
       else
          u_data_valid_25 <= '0';
       end if;
    end if;
  end process;
  
  process(clk) begin
    if (clk 'event and clk = '1') then
       if(u_data_valid_en_p3 = '1' and u_data_valid_en = '1') then 
          u_data_valid_20 <= '1';
       else
          u_data_valid_20 <= '0';
       end if;
    end if;
  end process;

  process (cas_lat_half(1), u_data_valid_25, u_data_valid_20)
  begin  -- process
    if (cas_lat_half(1) = '1') then
      u_data_valid <= u_data_valid_25;
    else
      u_data_valid <= u_data_valid_20;    
    end if;
  end process;

  -- IO buffers to interface with DDR SDRAM
  g5: for i in 0 to DDR_DATA_MSB generate
    ddr_dq_io: IOBUF_SSTL2_II port map(I=>ddr_dq_o(i), IO=>ddr_dq(i),
                                       O=>ddr_dq_tmp(i), T=>ddr_dq_t(i));
  end generate;
  ddr_dq_i <= ddr_dq_tmp; 

end rtl;
